/* 
   Wipe V1.0-- Overwrite and delete files.  S. 2/3/96

   This program attempts to securely delete any file specified on the
   command line, by overwriting it with other data multiple times.  It
   also overwrites the name using mktemp and rename.  

   USE THIS PROGRAM AT YOUR OWN RISK: THERE IS NO WARRANTY OF ANY KIND.
   This program is in the public domain.  It may be used without restriction.

   The overwrite passes are controlled by the passes[] array.  The
   commented passes[] array approximates the 'government' mode of
   Norton's wf.  It always wipes in multiples of BUFSIZE, so set this
   to >= the cluster size to wipe the unused bytes past the end of the
   file.

   This can simulate a free space disk wipe (as root):
   cat /dev/zero > temp; wipe temp
*/

#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>

/* Pass types:
    0: all 0's
   -1: all 1's
    1: even bits = 1
    2: odd bits = 1
    3: pseudorandom
*/
int passes[] = { 3, -1, 0, -1, 0, 2, 1, 2, 1, 3};
/* int passes[] = { -1, 0, -1, 0, -1, 0, 3 }; */

#define NUMPASSES (sizeof(passes)/sizeof(passes[0]))
 
#define BUFSIZE 4096
char outbuf[BUFSIZE];

char nameset[]="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789_+=%@#.";
#define NUMNSET (strlen(nameset))

/* Get file size. */
long filesize(int fd)
{
  struct stat fs;
  
  if(fstat(fd, &fs))
    {perror("Can't fstat file");}
  return (long)fs.st_size;
}

/* init randomizer. */
void initrandom( void )
{
  struct timeval tv;
  struct timezone tz;
  unsigned int seed;

  gettimeofday(&tv,&tz);
  seed = tv.tv_sec^(tv.tv_usec<<11);
#ifdef DEBUG
  printf("\tInitializing random to: %u\n", seed);
#endif
  srandom(seed);
}

/* Fill the output buffer. */
void filltable( int type)
{
  int c;
  switch( type )
    {
    case 0:  for(c = 0; c < BUFSIZE; c ++) outbuf[c] = 0; break;
    case -1: for(c = 0; c < BUFSIZE; c ++) outbuf[c] = -1; break;
    case 1:  for(c = 0; c < BUFSIZE; c ++) outbuf[c] = 0x55; break;
    case 2:  for(c = 0; c < BUFSIZE; c ++) outbuf[c] = 0xaa; break;
    case 3:  for(c = 0; c < BUFSIZE; c ++) outbuf[c] = random(); break;
    }
#ifdef DEBUG
  printf("\t\tCheck: %d\n", (int)outbuf[0]);
#endif
}

/* Make a random file character. */
char randchar( void )
{
  long int i,j;
  i = RAND_MAX/NUMNSET;
  i *= NUMNSET;
  while((j = random()) >= i);
  return nameset[j%NUMNSET];
}

/* Make a new name for a file. */
char *wipename(char *oldname)
{
  char *out, *c;
  if(!(out = malloc(strlen(oldname)+7)))
    {perror("Can't malloc");exit(-1);}
  strcpy(out, oldname);
  if(c = strrchr(out, '/')) c ++; else c = out;
  for(;strlen(c) > 6;c ++)
    *c = randchar();
  strcpy(c, "XXXXXX");
  if(!(out = mktemp( out )))
    {perror("Can't mktemp");exit(-1);}
#ifdef DEBUG
  printf("\tRenaming '%s' to '%s'\n", oldname, out);
#endif 
  return out;
}

int main( int argc, char *argv[] )
{
  int todel, i, pc, o = -1;
  long c, len;
  char *nn;
  
  for(i = 1; i < argc; i ++)
    {
      initrandom();
      if((todel = open(argv[i], O_RDWR)) == -1)
	{
	  printf("Can't open file: %s!\n", argv[i]);
	  continue;
	}
      len = filesize(todel);
      for(pc = 0; pc < NUMPASSES; pc ++)
	{
#ifdef DEBUG
	  printf("File: %s, Pass number: %d, Type: %s\n", argv[i], pc, 
		 (passes[pc]==-1)?"Ones":
		 ((passes[pc]==0)?"Zeros":
		  ((passes[pc]==1)?"Even":
		   ((passes[pc]==2)?"Odd":
		    ((passes[pc]==3)?"Random":"????")))));
#endif
	  filltable(passes[pc]);
	  for(c = 0; c < len; c += BUFSIZE)
	    {
	      if(passes[pc] == 3)
		filltable(3);
	      if(write(todel,outbuf,BUFSIZE) == -1)
		{printf("Write returns error, loc %ld: %s\n", c, strerror(errno));exit(-1);}
	    }
	  fsync(todel);
	  lseek(todel, 0, SEEK_SET);
	}
      close(todel);
      if(open(nn=wipename(argv[i]),O_RDONLY) != -1)
	{printf("Bad temp name generation: '%s'!\n",nn);exit(-1);}
      if(rename(argv[i], nn))
	{perror("Can't rename");exit(-1);}
#ifndef DEBUG
      if(unlink(nn))
	{perror("Can't unlink file");exit(-1);}
#endif
      free(nn);
      o = 0;
    }
return o;
}
